@@ -1,53 +1,38 @@
+from __future__ import annotations
+
 import datetime
 import logging
 import urllib.parse as urlparse
 import xml.etree.ElementTree as etree
+from abc import abstractmethod
 from inspect import getfullargspec
 from inspect import signature
 
-import requests
-from requests.exceptions import HTTPError
+import aiohttp
+import aiostream
+
+from vdirsyncer.exceptions import Error
+from vdirsyncer.vobject import Item
 
 from .. import exceptions
 from .. import http
 from .. import utils
+from ..http import USERAGENT
 from ..http import prepare_auth
 from ..http import prepare_client_cert
 from ..http import prepare_verify
-from ..http import USERAGENT
-from ..vobject import Item
-from .base import normalize_meta_value
 from .base import Storage
-
+from .base import normalize_meta_value
 
 dav_logger = logging.getLogger(__name__)
 
 CALDAV_DT_FORMAT = "%Y%m%dT%H%M%SZ"
 
 
-def _generate_path_reserved_chars():
-    for x in "/?#[]!$&'()*+,;":
-        x = urlparse.quote(x, "")
-        yield x.upper()
-        yield x.lower()
-
-
-_path_reserved_chars = frozenset(_generate_path_reserved_chars())
-del _generate_path_reserved_chars
-
-
-def _contains_quoted_reserved_chars(x):
-    for y in _path_reserved_chars:
-        if y in x:
-            dav_logger.debug(f"Unsafe character: {y!r}")
-            return True
-    return False
-
-
-def _assert_multistatus_success(r):
+async def _assert_multistatus_success(r):
     # Xandikos returns a multistatus on PUT.
     try:
-        root = _parse_xml(r.content)
+        root = _parse_xml(await r.content.read())
     except InvalidXMLResponse:
         return
     for status in root.findall(".//{DAV:}status"):
@@ -57,12 +42,11 @@ def _assert_multistatus_success(r):
         except (ValueError, IndexError):
             continue
         if st < 200 or st >= 400:
-            raise HTTPError(f"Server error: {st}")
+            raise Error(f"Server error: {st}")
 
 
 def _normalize_href(base, href):
-    """Normalize the href to be a path only relative to hostname and
-    schema."""
+    """Normalize the href to be a path only relative to hostname and schema."""
     orig_href = href
     if not href:
         raise ValueError(href)
@@ -70,22 +54,15 @@ def _normalize_href(base, href):
     x = urlparse.urljoin(base, href)
     x = urlparse.urlsplit(x).path
 
-    # Encoding issues:
-    # - https://github.com/owncloud/contacts/issues/581
-    # - https://github.com/Kozea/Radicale/issues/298
-    old_x = None
-    while old_x is None or x != old_x:
-        if _contains_quoted_reserved_chars(x):
-            break
-        old_x = x
-        x = urlparse.unquote(x)
-
-    x = urlparse.quote(x, "/@%:")
+    # We unquote and quote again, but want to make sure we
+    # keep around the "@" character.
+    x = urlparse.unquote(x)
+    x = urlparse.quote(x, "/@")
 
     if orig_href == x:
         dav_logger.debug(f"Already normalized: {x!r}")
     else:
-        dav_logger.debug("Normalized URL from {!r} to {!r}".format(orig_href, x))
+        dav_logger.debug(f"Normalized URL from {orig_href!r} to {x!r}")
 
     return x
 
@@ -115,8 +92,7 @@ def _parse_xml(content):
         return etree.XML(_clean_body(content))
     except etree.ParseError as e:
         raise InvalidXMLResponse(
-            "Invalid XML encountered: {}\n"
-            "Double-check the URLs in your config.".format(e)
+            f"Invalid XML encountered: {e}\nDouble-check the URLs in your config."
         )
 
 
@@ -144,11 +120,31 @@ def _fuzzy_matches_mimetype(strict, weak):
 
 
 class Discover:
-    _namespace = None
-    _resourcetype = None
-    _homeset_xml = None
-    _homeset_tag = None
-    _well_known_uri = None
+    @property
+    @abstractmethod
+    def _namespace(self) -> str:
+        pass
+
+    @property
+    @abstractmethod
+    def _resourcetype(self) -> str | None:
+        pass
+
+    @property
+    @abstractmethod
+    def _homeset_xml(self) -> bytes:
+        pass
+
+    @property
+    @abstractmethod
+    def _homeset_tag(self) -> str:
+        pass
+
+    @property
+    @abstractmethod
+    def _well_known_uri(self) -> str:
+        pass
+
     _collection_xml = b"""
     <propfind xmlns="DAV:">
         <prop>
@@ -169,14 +165,14 @@ class Discover:
         _, collection = url.rstrip("/").rsplit("/", 1)
         return urlparse.unquote(collection)
 
-    def find_principal(self):
+    async def find_principal(self):
         try:
-            return self._find_principal_impl("")
-        except (HTTPError, exceptions.Error):
+            return await self._find_principal_impl("")
+        except (aiohttp.ClientResponseError, exceptions.Error):
             dav_logger.debug("Trying out well-known URI")
-            return self._find_principal_impl(self._well_known_uri)
+            return await self._find_principal_impl(self._well_known_uri)
 
-    def _find_principal_impl(self, url):
+    async def _find_principal_impl(self, url) -> str:
         headers = self.session.get_default_headers()
         headers["Depth"] = "0"
         body = b"""
@@ -187,48 +183,54 @@ class Discover:
         </propfind>
         """
 
-        response = self.session.request("PROPFIND", url, headers=headers, data=body)
+        response = await self.session.request(
+            "PROPFIND",
+            url,
+            headers=headers,
+            data=body,
+        )
 
-        root = _parse_xml(response.content)
+        root = _parse_xml(await response.content.read())
         rv = root.find(".//{DAV:}current-user-principal/{DAV:}href")
         if rv is None:
             # This is for servers that don't support current-user-principal
             # E.g. Synology NAS
             # See https://github.com/pimutils/vdirsyncer/issues/498
             dav_logger.debug(
-                "No current-user-principal returned, re-using URL {}".format(
-                    response.url
-                )
+                f"No current-user-principal returned, re-using URL {response.url}"
             )
-            return response.url
-        return urlparse.urljoin(response.url, rv.text).rstrip("/") + "/"
+            return response.url.human_repr()
+        return urlparse.urljoin(str(response.url), rv.text).rstrip("/") + "/"
 
-    def find_home(self):
-        url = self.find_principal()
+    async def find_home(self):
+        url = await self.find_principal()
         headers = self.session.get_default_headers()
         headers["Depth"] = "0"
-        response = self.session.request(
+        response = await self.session.request(
             "PROPFIND", url, headers=headers, data=self._homeset_xml
         )
 
-        root = etree.fromstring(response.content)
+        root = etree.fromstring(await response.content.read())
         # Better don't do string formatting here, because of XML namespaces
         rv = root.find(".//" + self._homeset_tag + "/{DAV:}href")
         if rv is None:
             raise InvalidXMLResponse("Couldn't find home-set.")
-        return urlparse.urljoin(response.url, rv.text).rstrip("/") + "/"
+        return urlparse.urljoin(str(response.url), rv.text).rstrip("/") + "/"
 
-    def find_collections(self):
+    async def find_collections(self):
         rv = None
         try:
-            rv = list(self._find_collections_impl(""))
-        except (HTTPError, exceptions.Error):
+            rv = await aiostream.stream.list(self._find_collections_impl(""))
+        except (aiohttp.ClientResponseError, exceptions.Error):
             pass
 
         if rv:
             return rv
+
         dav_logger.debug("Given URL is not a homeset URL")
-        return self._find_collections_impl(self.find_home())
+        return await aiostream.stream.list(
+            self._find_collections_impl(await self.find_home())
+        )
 
     def _check_collection_resource_type(self, response):
         if self._resourcetype is None:
@@ -245,13 +247,13 @@ class Discover:
             return False
         return True
 
-    def _find_collections_impl(self, url):
+    async def _find_collections_impl(self, url):
         headers = self.session.get_default_headers()
         headers["Depth"] = "1"
-        r = self.session.request(
+        r = await self.session.request(
             "PROPFIND", url, headers=headers, data=self._collection_xml
         )
-        root = _parse_xml(r.content)
+        root = _parse_xml(await r.content.read())
         done = set()
         for response in root.findall("{DAV:}response"):
             if not self._check_collection_resource_type(response):
@@ -260,33 +262,33 @@ class Discover:
             href = response.find("{DAV:}href")
             if href is None:
                 raise InvalidXMLResponse("Missing href tag for collection " "props.")
-            href = urlparse.urljoin(r.url, href.text)
+            href = urlparse.urljoin(str(r.url), href.text)
             if href not in done:
                 done.add(href)
                 yield {"href": href}
 
-    def discover(self):
-        for c in self.find_collections():
+    async def discover(self):
+        for c in await self.find_collections():
             url = c["href"]
             collection = self._get_collection_from_url(url)
             storage_args = dict(self.kwargs)
             storage_args.update({"url": url, "collection": collection})
             yield storage_args
 
-    def create(self, collection):
+    async def create(self, collection):
         if collection is None:
             collection = self._get_collection_from_url(self.kwargs["url"])
 
-        for c in self.discover():
+        async for c in self.discover():
             if c["collection"] == collection:
                 return c
 
-        home = self.find_home()
+        home = await self.find_home()
         url = urlparse.urljoin(home, urlparse.quote(collection, "/@"))
 
         try:
-            url = self._create_collection_impl(url)
-        except HTTPError as e:
+            url = await self._create_collection_impl(url)
+        except (aiohttp.ClientResponseError, Error) as e:
             raise NotImplementedError(e)
         else:
             rv = dict(self.kwargs)
@@ -294,7 +296,7 @@ class Discover:
             rv["url"] = url
             return rv
 
-    def _create_collection_impl(self, url):
+    async def _create_collection_impl(self, url):
         data = """<?xml version="1.0" encoding="utf-8" ?>
             <mkcol xmlns="DAV:">
                 <set>
@@ -312,13 +314,13 @@ class Discover:
             "utf-8"
         )
 
-        response = self.session.request(
+        response = await self.session.request(
             "MKCOL",
             url,
             data=data,
             headers=self.session.get_default_headers(),
         )
-        return response.url
+        return str(response.url)
 
 
 class CalDiscover(Discover):
@@ -337,7 +339,7 @@ class CalDiscover(Discover):
 
 class CardDiscover(Discover):
     _namespace = "urn:ietf:params:xml:ns:carddav"
-    _resourcetype = "{%s}addressbook" % _namespace
+    _resourcetype: str | None = "{%s}addressbook" % _namespace
     _homeset_xml = b"""
     <propfind xmlns="DAV:" xmlns:c="urn:ietf:params:xml:ns:carddav">
         <prop>
@@ -350,14 +352,18 @@ class CardDiscover(Discover):
 
 
 class DAVSession:
-    """
-    A helper class to connect to DAV servers.
-    """
+    """A helper class to connect to DAV servers."""
+
+    connector: aiohttp.BaseConnector
 
     @classmethod
     def init_and_remaining_args(cls, **kwargs):
+        def is_arg(k):
+            """Return true if ``k`` is an argument of ``cls.__init__``."""
+            return k in argspec.args or k in argspec.kwonlyargs
+
         argspec = getfullargspec(cls.__init__)
-        self_args, remainder = utils.split_dict(kwargs, argspec.args.__contains__)
+        self_args, remainder = utils.split_dict(kwargs, is_arg)
 
         return cls(**self_args), remainder
 
@@ -366,35 +372,56 @@ class DAVSession:
         url,
         username="",
         password="",
-        verify=True,
+        verify=None,
         auth=None,
         useragent=USERAGENT,
         verify_fingerprint=None,
         auth_cert=None,
+        *,
+        connector: aiohttp.BaseConnector,
     ):
         self._settings = {
             "cert": prepare_client_cert(auth_cert),
-            "auth": prepare_auth(auth, username, password),
         }
-        self._settings.update(prepare_verify(verify, verify_fingerprint))
+        auth = prepare_auth(auth, username, password)
+        if auth:
+            self._settings["auth"] = auth
+
+        ssl = prepare_verify(verify, verify_fingerprint)
+        if ssl:
+            self._settings["ssl"] = ssl
 
         self.useragent = useragent
         self.url = url.rstrip("/") + "/"
-
-        self._session = requests.session()
+        self.connector = connector
 
     @utils.cached_property
     def parsed_url(self):
         return urlparse.urlparse(self.url)
 
-    def request(self, method, path, **kwargs):
+    async def request(self, method, path, **kwargs):
         url = self.url
         if path:
-            url = urlparse.urljoin(self.url, path)
+            url = urlparse.urljoin(str(self.url), path)
 
         more = dict(self._settings)
         more.update(kwargs)
-        return http.request(method, url, session=self._session, **more)
+
+        # XXX: This is a temporary hack to pin-point bad refactoring.
+        assert self.connector is not None
+        async with self._session as session:
+            return await http.request(method, url, session=session, **more)
+
+    @property
+    def _session(self):
+        """Return a new session for requests."""
+
+        return aiohttp.ClientSession(
+            connector=self.connector,
+            connector_owner=False,
+            trust_env=True,
+            # TODO use `raise_for_status=true`, though this needs traces first,
+        )
 
     def get_default_headers(self):
         return {
@@ -405,45 +432,64 @@ class DAVSession:
 
 class DAVStorage(Storage):
     # the file extension of items. Useful for testing against radicale.
-    fileext = None
+    fileext: str
     # mimetype of items
-    item_mimetype = None
-    # XML to use when fetching multiple hrefs.
-    get_multi_template = None
-    # The LXML query for extracting results in get_multi
-    get_multi_data_query = None
-    # The Discover subclass to use
-    discovery_class = None
+    item_mimetype: str
+
+    @property
+    @abstractmethod
+    def get_multi_template(self) -> str:
+        """XML to use when fetching multiple hrefs."""
+
+    @property
+    @abstractmethod
+    def get_multi_data_query(self) -> str:
+        """LXML query for extracting results in get_multi."""
+
+    @property
+    @abstractmethod
+    def discovery_class(self) -> type[Discover]:
+        """Discover subclass to use."""
+
     # The DAVSession class to use
     session_class = DAVSession
 
-    _repr_attributes = ("username", "url")
+    connector: aiohttp.TCPConnector
+
+    _repr_attributes = ["username", "url"]
 
     _property_table = {
         "displayname": ("displayname", "DAV:"),
     }
 
-    def __init__(self, **kwargs):
+    def __init__(self, *, connector, **kwargs):
         # defined for _repr_attributes
         self.username = kwargs.get("username")
         self.url = kwargs.get("url")
+        self.connector = connector
 
-        self.session, kwargs = self.session_class.init_and_remaining_args(**kwargs)
+        self.session, kwargs = self.session_class.init_and_remaining_args(
+            connector=connector,
+            **kwargs,
+        )
         super().__init__(**kwargs)
 
-    __init__.__signature__ = signature(session_class.__init__)
+    __init__.__signature__ = signature(session_class.__init__)  # type: ignore
+    # See  https://github.com/python/mypy/issues/5958
 
     @classmethod
-    def discover(cls, **kwargs):
+    async def discover(cls, **kwargs):
         session, _ = cls.session_class.init_and_remaining_args(**kwargs)
         d = cls.discovery_class(session, kwargs)
-        return d.discover()
+
+        async for collection in d.discover():
+            yield collection
 
     @classmethod
-    def create_collection(cls, collection, **kwargs):
+    async def create_collection(cls, collection, **kwargs):
         session, _ = cls.session_class.init_and_remaining_args(**kwargs)
         d = cls.discovery_class(session, kwargs)
-        return d.create(collection)
+        return await d.create(collection)
 
     def _normalize_href(self, *args, **kwargs):
         return _normalize_href(self.session.url, *args, **kwargs)
@@ -455,57 +501,61 @@ class DAVStorage(Storage):
     def _is_item_mimetype(self, mimetype):
         return _fuzzy_matches_mimetype(self.item_mimetype, mimetype)
 
-    def get(self, href):
-        ((actual_href, item, etag),) = self.get_multi([href])
+    async def get(self, href: str):
+        ((actual_href, item, etag),) = await aiostream.stream.list(
+            self.get_multi([href])
+        )
         assert href == actual_href
         return item, etag
 
-    def get_multi(self, hrefs):
+    async def get_multi(self, hrefs):
         hrefs = set(hrefs)
         href_xml = []
         for href in hrefs:
             if href != self._normalize_href(href):
                 raise exceptions.NotFoundError(href)
             href_xml.append(f"<href>{href}</href>")
-        if not href_xml:
-            return ()
+        if href_xml:
+            data = self.get_multi_template.format(hrefs="\n".join(href_xml)).encode(
+                "utf-8"
+            )
+            response = await self.session.request(
+                "REPORT", "", data=data, headers=self.session.get_default_headers()
+            )
+            root = _parse_xml(
+                await response.content.read()
+            )  # etree only can handle bytes
+            rv = []
+            hrefs_left = set(hrefs)
+            for href, etag, prop in self._parse_prop_responses(root):
+                raw = prop.find(self.get_multi_data_query)
+                if raw is None:
+                    dav_logger.warning(f"Skipping {href}, the item content is missing.")
+                    continue
 
-        data = self.get_multi_template.format(hrefs="\n".join(href_xml)).encode("utf-8")
-        response = self.session.request(
-            "REPORT", "", data=data, headers=self.session.get_default_headers()
-        )
-        root = _parse_xml(response.content)  # etree only can handle bytes
-        rv = []
-        hrefs_left = set(hrefs)
-        for href, etag, prop in self._parse_prop_responses(root):
-            raw = prop.find(self.get_multi_data_query)
-            if raw is None:
-                dav_logger.warning(
-                    "Skipping {}, the item content is missing.".format(href)
-                )
-                continue
+                raw = raw.text or ""
 
-            raw = raw.text or ""
+                if isinstance(raw, bytes):
+                    raw = raw.decode(response.encoding)
+                if isinstance(etag, bytes):
+                    etag = etag.decode(response.encoding)
 
-            if isinstance(raw, bytes):
-                raw = raw.decode(response.encoding)
-            if isinstance(etag, bytes):
-                etag = etag.decode(response.encoding)
-
-            try:
-                hrefs_left.remove(href)
-            except KeyError:
-                if href in hrefs:
-                    dav_logger.warning("Server sent item twice: {}".format(href))
+                try:
+                    hrefs_left.remove(href)
+                except KeyError:
+                    if href in hrefs:
+                        dav_logger.warning(f"Server sent item twice: {href}")
+                    else:
+                        dav_logger.warning(f"Server sent unsolicited item: {href}")
                 else:
-                    dav_logger.warning("Server sent unsolicited item: {}".format(href))
-            else:
-                rv.append((href, Item(raw), etag))
-        for href in hrefs_left:
-            raise exceptions.NotFoundError(href)
-        return rv
+                    rv.append((href, Item(raw), etag))
+            for href in hrefs_left:
+                raise exceptions.NotFoundError(href)
 
-    def _put(self, href, item, etag):
+            for href, item, etag in rv:
+                yield href, item, etag
+
+    async def _put(self, href, item, etag):
         headers = self.session.get_default_headers()
         headers["Content-Type"] = self.item_mimetype
         if etag is None:
@@ -513,11 +563,11 @@ class DAVStorage(Storage):
         else:
             headers["If-Match"] = etag
 
-        response = self.session.request(
+        response = await self.session.request(
             "PUT", href, data=item.raw.encode("utf-8"), headers=headers
         )
 
-        _assert_multistatus_success(response)
+        await _assert_multistatus_success(response)
 
         # The server may not return an etag under certain conditions:
         #
@@ -534,25 +584,28 @@ class DAVStorage(Storage):
         # In such cases we return a constant etag. The next synchronization
         # will then detect an etag change and will download the new item.
         etag = response.headers.get("etag", None)
-        href = self._normalize_href(response.url)
+        href = self._normalize_href(str(response.url))
         return href, etag
 
-    def update(self, href, item, etag):
+    async def update(self, href, item, etag):
         if etag is None:
             raise ValueError("etag must be given and must not be None.")
-        href, etag = self._put(self._normalize_href(href), item, etag)
+        href, etag = await self._put(self._normalize_href(href), item, etag)
         return etag
 
-    def upload(self, item):
+    async def upload(self, item: Item):
         href = self._get_href(item)
-        return self._put(href, item, None)
+        rv = await self._put(href, item, None)
+        return rv
 
-    def delete(self, href, etag):
+    async def delete(self, href, etag):
         href = self._normalize_href(href)
         headers = self.session.get_default_headers()
-        headers.update({"If-Match": etag})
+        if etag:  # baikal doesn't give us an etag.
+            dav_logger.warning("Deleting an item with no etag.")
+            headers.update({"If-Match": etag})
 
-        self.session.request("DELETE", href, headers=headers)
+        await self.session.request("DELETE", href, headers=headers)
 
     def _parse_prop_responses(self, root, handled_hrefs=None):
         if handled_hrefs is None:
@@ -571,12 +624,12 @@ class DAVStorage(Storage):
                 #   https://github.com/pimutils/vdirsyncer/issues/88
                 # - Davmail
                 #   https://github.com/pimutils/vdirsyncer/issues/144
-                dav_logger.warning("Skipping identical href: {!r}".format(href))
+                dav_logger.warning(f"Skipping identical href: {href!r}")
                 continue
 
             props = response.findall("{DAV:}propstat/{DAV:}prop")
             if props is None or not len(props):
-                dav_logger.debug("Skipping {!r}, properties are missing.".format(href))
+                dav_logger.debug(f"Skipping {href!r}, properties are missing.")
                 continue
             else:
                 props = _merge_xml(props)
@@ -587,24 +640,20 @@ class DAVStorage(Storage):
 
             etag = getattr(props.find("{DAV:}getetag"), "text", "")
             if not etag:
-                dav_logger.debug(
-                    "Skipping {!r}, etag property is missing.".format(href)
-                )
+                dav_logger.debug(f"Skipping {href!r}, etag property is missing.")
                 continue
 
             contenttype = getattr(props.find("{DAV:}getcontenttype"), "text", None)
             if not self._is_item_mimetype(contenttype):
                 dav_logger.debug(
-                    "Skipping {!r}, {!r} != {!r}.".format(
-                        href, contenttype, self.item_mimetype
-                    )
+                    f"Skipping {href!r}, {contenttype!r} != {self.item_mimetype!r}."
                 )
                 continue
 
             handled_hrefs.add(href)
             yield href, etag, props
 
-    def list(self):
+    async def list(self):
         headers = self.session.get_default_headers()
         headers["Depth"] = "1"
 
@@ -620,71 +669,86 @@ class DAVStorage(Storage):
 
         # We use a PROPFIND request instead of addressbook-query due to issues
         # with Zimbra. See https://github.com/pimutils/vdirsyncer/issues/83
-        response = self.session.request("PROPFIND", "", data=data, headers=headers)
-        root = _parse_xml(response.content)
+        response = await self.session.request(
+            "PROPFIND",
+            "",
+            data=data,
+            headers=headers,
+        )
+        root = _parse_xml(await response.content.read())
 
         rv = self._parse_prop_responses(root)
         for href, etag, _prop in rv:
             yield href, etag
 
-    def get_meta(self, key):
+    async def get_meta(self, key) -> str | None:
         try:
             tagname, namespace = self._property_table[key]
         except KeyError:
-            raise exceptions.UnsupportedMetadataError()
+            raise exceptions.UnsupportedMetadataError
 
         xpath = f"{{{namespace}}}{tagname}"
-        data = """<?xml version="1.0" encoding="utf-8" ?>
+        body = f"""<?xml version="1.0" encoding="utf-8" ?>
             <propfind xmlns="DAV:">
                 <prop>
-                    {}
+                    {etree.tostring(etree.Element(xpath), encoding="unicode")}
                 </prop>
             </propfind>
-        """.format(
-            etree.tostring(etree.Element(xpath), encoding="unicode")
-        ).encode(
-            "utf-8"
-        )
+        """
+        data = body.encode("utf-8")
 
         headers = self.session.get_default_headers()
         headers["Depth"] = "0"
 
-        response = self.session.request("PROPFIND", "", data=data, headers=headers)
+        response = await self.session.request(
+            "PROPFIND",
+            "",
+            data=data,
+            headers=headers,
+        )
 
-        root = _parse_xml(response.content)
+        root = _parse_xml(await response.content.read())
 
         for prop in root.findall(".//" + xpath):
             text = normalize_meta_value(getattr(prop, "text", None))
             if text:
                 return text
-        return ""
+        return None
 
-    def set_meta(self, key, value):
+    async def set_meta(self, key, value):
         try:
             tagname, namespace = self._property_table[key]
         except KeyError:
-            raise exceptions.UnsupportedMetadataError()
+            raise exceptions.UnsupportedMetadataError
 
         lxml_selector = f"{{{namespace}}}{tagname}"
         element = etree.Element(lxml_selector)
-        element.text = normalize_meta_value(value)
+        if value is None:
+            action = "remove"
+        else:
+            element.text = normalize_meta_value(value)
+            action = "set"
 
         data = """<?xml version="1.0" encoding="utf-8" ?>
             <propertyupdate xmlns="DAV:">
-                <set>
+                <{action}>
                     <prop>
                         {}
                     </prop>
-                </set>
+                </{action}>
             </propertyupdate>
         """.format(
-            etree.tostring(element, encoding="unicode")
+            etree.tostring(element, encoding="unicode"),
+            action=action,
         ).encode(
             "utf-8"
         )
 
-        self.session.request(
-            "PROPPATCH", "", data=data, headers=self.session.get_default_headers()
+        await self.session.request(
+            "PROPPATCH",
+            "",
+            data=data,
+            headers=self.session.get_default_headers(),
         )
 
         # XXX: Response content is currently ignored. Though exceptions are
@@ -719,6 +783,8 @@ class CalDAVStorage(DAVStorage):
     _property_table.update(
         {
             "color": ("calendar-color", "http://apple.com/ns/ical/"),
+            "description": ("calendar-description", "urn:ietf:params:xml:ns:caldav"),
+            "order": ("calendar-order", "http://apple.com/ns/ical/"),
         }
     )
 
@@ -760,9 +826,7 @@ class CalDAVStorage(DAVStorage):
                 start = start.strftime(CALDAV_DT_FORMAT)
                 end = end.strftime(CALDAV_DT_FORMAT)
 
-                timefilter = '<C:time-range start="{start}" end="{end}"/>'.format(
-                    start=start, end=end
-                )
+                timefilter = f'<C:time-range start="{start}" end="{end}"/>'
             else:
                 timefilter = ""
 
@@ -774,7 +838,7 @@ class CalDAVStorage(DAVStorage):
                     ("VTODO", "VEVENT"), start, end
                 )
 
-    def list(self):
+    async def list(self):
         caldavfilters = list(
             self._get_list_filters(self.item_types, self.start_date, self.end_date)
         )
@@ -786,7 +850,8 @@ class CalDAVStorage(DAVStorage):
             # instead?
             #
             # See https://github.com/dmfs/tasks/issues/118 for backstory.
-            yield from DAVStorage.list(self)
+            async for href, etag in super().list():
+                yield href, etag
 
         data = """<?xml version="1.0" encoding="utf-8" ?>
             <C:calendar-query xmlns="DAV:"
@@ -811,8 +876,13 @@ class CalDAVStorage(DAVStorage):
 
         for caldavfilter in caldavfilters:
             xml = data.format(caldavfilter=caldavfilter).encode("utf-8")
-            response = self.session.request("REPORT", "", data=xml, headers=headers)
-            root = _parse_xml(response.content)
+            response = await self.session.request(
+                "REPORT",
+                "",
+                data=xml,
+                headers=headers,
+            )
+            root = _parse_xml(await response.content.read())
             rv = self._parse_prop_responses(root, handled_hrefs)
             for href, etag, _prop in rv:
                 yield href, etag
@@ -824,14 +894,31 @@ class CardDAVStorage(DAVStorage):
     item_mimetype = "text/vcard"
     discovery_class = CardDiscover
 
-    get_multi_template = """<?xml version="1.0" encoding="utf-8" ?>
+    def __init__(self, *args, use_vcard_4=False, **kwargs):
+        self.use_vcard_4 = use_vcard_4
+        super().__init__(*args, **kwargs)
+
+    @property
+    def get_multi_template(self):
+        ct = 'Content-Type="text/vcard" version="4.0"' if self.use_vcard_4 else ""
+        return f"""<?xml version="1.0" encoding="utf-8" ?>
             <C:addressbook-multiget xmlns="DAV:"
                     xmlns:C="urn:ietf:params:xml:ns:carddav">
                 <prop>
                     <getetag/>
-                    <C:address-data/>
+                    <C:address-data {ct}/>
                 </prop>
-                {hrefs}
+                {{hrefs}}
             </C:addressbook-multiget>"""
 
     get_multi_data_query = "{urn:ietf:params:xml:ns:carddav}address-data"
+
+    _property_table = dict(DAVStorage._property_table)
+    _property_table.update(
+        {
+            "description": (
+                "addressbook-description",
+                "urn:ietf:params:xml:ns:carddav",
+            ),
+        }
+    )
