@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import errno
 import logging
 import os
@@ -11,16 +13,15 @@ from ..utils import expand_path
 from ..utils import generate_href
 from ..utils import get_etag_from_file
 from ..vobject import Item
-from .base import normalize_meta_value
 from .base import Storage
+from .base import normalize_meta_value
 
 logger = logging.getLogger(__name__)
 
 
 class FilesystemStorage(Storage):
-
     storage_name = "filesystem"
-    _repr_attributes = ("path",)
+    _repr_attributes = ["path"]
 
     def __init__(
         self,
@@ -28,8 +29,9 @@ class FilesystemStorage(Storage):
         fileext,
         encoding="utf-8",
         post_hook=None,
+        pre_deletion_hook=None,
         fileignoreext=".tmp",
-        **kwargs
+        **kwargs,
     ):
         super().__init__(**kwargs)
         path = expand_path(path)
@@ -39,9 +41,10 @@ class FilesystemStorage(Storage):
         self.fileext = fileext
         self.fileignoreext = fileignoreext
         self.post_hook = post_hook
+        self.pre_deletion_hook = pre_deletion_hook
 
     @classmethod
-    def discover(cls, path, **kwargs):
+    async def discover(cls, path, **kwargs):
         if kwargs.pop("collection", None) is not None:
             raise TypeError("collection argument must not be given.")
         path = expand_path(path)
@@ -67,7 +70,7 @@ class FilesystemStorage(Storage):
         return True
 
     @classmethod
-    def create_collection(cls, collection, **kwargs):
+    async def create_collection(cls, collection, **kwargs):
         kwargs = dict(kwargs)
         path = kwargs["path"]
 
@@ -86,7 +89,7 @@ class FilesystemStorage(Storage):
     def _get_href(self, ident):
         return generate_href(ident) + self.fileext
 
-    def list(self):
+    async def list(self):
         for fname in os.listdir(self.path):
             fpath = os.path.join(self.path, fname)
             if (
@@ -96,7 +99,7 @@ class FilesystemStorage(Storage):
             ):
                 yield fname, get_etag_from_file(fpath)
 
-    def get(self, href):
+    async def get(self, href):
         fpath = self._get_filepath(href)
         try:
             with open(fpath, "rb") as f:
@@ -107,7 +110,7 @@ class FilesystemStorage(Storage):
             else:
                 raise
 
-    def upload(self, item):
+    async def upload(self, item):
         if not isinstance(item.raw, str):
             raise TypeError("item.raw must be a unicode string.")
 
@@ -116,7 +119,7 @@ class FilesystemStorage(Storage):
             fpath, etag = self._upload_impl(item, href)
         except OSError as e:
             if e.errno in (errno.ENAMETOOLONG, errno.ENOENT):  # Unix  # Windows
-                logger.debug("UID as filename rejected, trying with random " "one.")
+                logger.debug("UID as filename rejected, trying with random one.")
                 # random href instead of UID-based
                 href = self._get_href(None)
                 fpath, etag = self._upload_impl(item, href)
@@ -139,7 +142,7 @@ class FilesystemStorage(Storage):
             else:
                 raise
 
-    def update(self, href, item, etag):
+    async def update(self, href, item, etag):
         fpath = self._get_filepath(href)
         if not os.path.exists(fpath):
             raise exceptions.NotFoundError(item.uid)
@@ -158,38 +161,54 @@ class FilesystemStorage(Storage):
             self._run_post_hook(fpath)
         return etag
 
-    def delete(self, href, etag):
+    async def delete(self, href, etag):
         fpath = self._get_filepath(href)
         if not os.path.isfile(fpath):
             raise exceptions.NotFoundError(href)
         actual_etag = get_etag_from_file(fpath)
         if etag != actual_etag:
             raise exceptions.WrongEtagError(etag, actual_etag)
+        if self.pre_deletion_hook:
+            self._run_pre_deletion_hook(fpath)
+
         os.remove(fpath)
 
     def _run_post_hook(self, fpath):
-        logger.info(
-            "Calling post_hook={} with argument={}".format(self.post_hook, fpath)
-        )
+        logger.info(f"Calling post_hook={self.post_hook} with argument={fpath}")
         try:
             subprocess.call([self.post_hook, fpath])
         except OSError as e:
-            logger.warning("Error executing external hook: {}".format(str(e)))
+            logger.warning(f"Error executing external hook: {str(e)}")
 
-    def get_meta(self, key):
+    def _run_pre_deletion_hook(self, fpath):
+        logger.info(
+            f"Calling pre_deletion_hook={self.pre_deletion_hook} with argument={fpath}"
+        )
+        try:
+            subprocess.call([self.pre_deletion_hook, fpath])
+        except OSError as e:
+            logger.warning(f"Error executing external hook: {str(e)}")
+
+    async def get_meta(self, key):
         fpath = os.path.join(self.path, key)
         try:
             with open(fpath, "rb") as f:
                 return normalize_meta_value(f.read().decode(self.encoding))
         except OSError as e:
             if e.errno == errno.ENOENT:
-                return ""
+                return None
             else:
                 raise
 
-    def set_meta(self, key, value):
+    async def set_meta(self, key, value):
         value = normalize_meta_value(value)
 
         fpath = os.path.join(self.path, key)
-        with atomic_write(fpath, mode="wb", overwrite=True) as f:
-            f.write(value.encode(self.encoding))
+        if value is None:
+            try:
+                os.remove(fpath)
+            except OSError:
+                pass
+        else:
+            with atomic_write(fpath, mode="wb", overwrite=True) as f:
+                f.write(value.encode(self.encoding))
