@@ -1,9 +1,12 @@
+from __future__ import annotations
+
 import collections
 import contextlib
 import functools
 import glob
 import logging
 import os
+from typing import Iterable
 
 from atomicwrites import atomic_write
 
@@ -11,6 +14,7 @@ from .. import exceptions
 from ..utils import checkfile
 from ..utils import expand_path
 from ..utils import get_etag_from_file
+from ..utils import uniq
 from ..vobject import Item
 from ..vobject import join_collection
 from ..vobject import split_collection
@@ -21,10 +25,12 @@ logger = logging.getLogger(__name__)
 
 def _writing_op(f):
     @functools.wraps(f)
-    def inner(self, *args, **kwargs):
+    async def inner(self, *args, **kwargs):
         if self._items is None or not self._at_once:
-            self.list()
-        rv = f(self, *args, **kwargs)
+            async for _ in self.list():
+                pass
+        assert self._items is not None
+        rv = await f(self, *args, **kwargs)
         if not self._at_once:
             self._write()
         return rv
@@ -34,7 +40,7 @@ def _writing_op(f):
 
 class SingleFileStorage(Storage):
     storage_name = "singlefile"
-    _repr_attributes = ("path",)
+    _repr_attributes = ["path"]
 
     _write_mode = "wb"
     _append_mode = "ab"
@@ -53,7 +59,7 @@ class SingleFileStorage(Storage):
         self._at_once = False
 
     @classmethod
-    def discover(cls, path, **kwargs):
+    async def discover(cls, path, **kwargs):
         if kwargs.pop("collection", None) is not None:
             raise TypeError("collection argument must not be given.")
 
@@ -63,7 +69,7 @@ class SingleFileStorage(Storage):
         except TypeError:
             # If not exactly one '%s' is present, we cannot discover
             # collections because we wouldn't know which name to assign.
-            raise NotImplementedError()
+            raise NotImplementedError
 
         placeholder_pos = path.index("%s")
 
@@ -81,7 +87,7 @@ class SingleFileStorage(Storage):
                 yield args
 
     @classmethod
-    def create_collection(cls, collection, **kwargs):
+    async def create_collection(cls, collection, **kwargs):
         path = os.path.abspath(expand_path(kwargs["path"]))
 
         if collection is not None:
@@ -97,7 +103,7 @@ class SingleFileStorage(Storage):
         kwargs["collection"] = collection
         return kwargs
 
-    def list(self):
+    async def list(self):
         self._items = collections.OrderedDict()
 
         try:
@@ -111,27 +117,33 @@ class SingleFileStorage(Storage):
                 raise OSError(e)
             text = None
 
-        if not text:
-            return ()
+        if text:
+            for item in split_collection(text):
+                item = Item(item)
+                etag = item.hash
+                href = item.ident
+                self._items[href] = item, etag
 
-        for item in split_collection(text):
-            item = Item(item)
-            etag = item.hash
-            self._items[item.ident] = item, etag
+                yield href, etag
 
-        return ((href, etag) for href, (item, etag) in self._items.items())
-
-    def get(self, href):
+    async def get(self, href):
         if self._items is None or not self._at_once:
-            self.list()
+            async for _ in self.list():
+                pass
 
         try:
             return self._items[href]
         except KeyError:
             raise exceptions.NotFoundError(href)
 
+    async def get_multi(self, hrefs: Iterable[str]):
+        async with self.at_once():
+            for href in uniq(hrefs):
+                item, etag = await self.get(href)
+                yield href, item, etag
+
     @_writing_op
-    def upload(self, item):
+    async def upload(self, item):
         href = item.ident
         if href in self._items:
             raise exceptions.AlreadyExistingError(existing_href=href)
@@ -140,7 +152,7 @@ class SingleFileStorage(Storage):
         return href, item.hash
 
     @_writing_op
-    def update(self, href, item, etag):
+    async def update(self, href, item, etag):
         if href not in self._items:
             raise exceptions.NotFoundError(href)
 
@@ -152,7 +164,7 @@ class SingleFileStorage(Storage):
         return item.hash
 
     @_writing_op
-    def delete(self, href, etag):
+    async def delete(self, href, etag):
         if href not in self._items:
             raise exceptions.NotFoundError(href)
 
@@ -167,11 +179,9 @@ class SingleFileStorage(Storage):
             self.path
         ):
             raise exceptions.PreconditionFailed(
-                (
-                    "Some other program modified the file {!r}. Re-run the "
-                    "synchronization and make sure absolutely no other program is "
-                    "writing into the same file."
-                ).format(self.path)
+                f"Some other program modified the file {self.path!r}. Re-run the "
+                "synchronization and make sure absolutely no other program is "
+                "writing into the same file."
             )
         text = join_collection(item.raw for item, etag in self._items.values())
         try:
@@ -181,9 +191,10 @@ class SingleFileStorage(Storage):
             self._items = None
             self._last_etag = None
 
-    @contextlib.contextmanager
-    def at_once(self):
-        self.list()
+    @contextlib.asynccontextmanager
+    async def at_once(self):
+        async for _ in self.list():
+            pass
         self._at_once = True
         try:
             yield self
