@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import hashlib
 from itertools import chain
 from itertools import tee
@@ -5,7 +7,6 @@ from itertools import tee
 from .utils import cached_property
 from .utils import uniq
 
-
 IGNORE_PROPS = (
     # PRODID is changed by radicale for some reason after upload
     "PRODID",
@@ -35,7 +36,6 @@ IGNORE_PROPS = (
 
 
 class Item:
-
     """Immutable wrapper class for VCALENDAR (VEVENT, VTODO) and
     VCARD"""
 
@@ -169,7 +169,7 @@ def _split_collection_impl(item, main, inline, items, ungrouped_items):
         for subitem in item.subcomponents:
             _split_collection_impl(subitem, item, inline, items, ungrouped_items)
     else:
-        raise ValueError("Unknown component: {}".format(item.name))
+        raise ValueError(f"Unknown component: {item.name}")
 
 
 _default_join_wrappers = {
@@ -280,6 +280,12 @@ class _Component:
                     stack.append(cls(c_name, [], []))
                 elif line.startswith("END:"):
                     component = stack.pop()
+                    c_name = line[len("END:") :].strip().upper()
+                    if c_name != component.name:
+                        raise ValueError(
+                            f"Got END:{c_name}, expected END:{component.name}"
+                            + f" at line {_i + 1}"
+                        )
                     if stack:
                         stack[-1].subcomponents.append(component)
                     else:
@@ -288,12 +294,17 @@ class _Component:
                     if line.strip():
                         stack[-1].props.append(line)
         except IndexError:
-            raise ValueError("Parsing error at line {}".format(_i + 1))
+            raise ValueError(f"Parsing error at line {_i + 1}")
+
+        if len(stack) > 0:
+            raise ValueError(
+                f"Missing END for component(s): {', '.join(c.name for c in stack)}"
+            )
 
         if multiple:
             return rv
         elif len(rv) != 1:
-            raise ValueError("Found {} components, expected one.".format(len(rv)))
+            raise ValueError(f"Found {len(rv)} components, expected one.")
         else:
             return rv[0]
 
@@ -353,7 +364,7 @@ class _Component:
                 rv = line[len(prefix_with_params) :].split(":", 1)[-1]
                 break
         else:
-            raise KeyError()
+            raise KeyError
 
         for line in iterlines:
             if line.startswith((" ", "\t")):
